/***************************************************************************
  par UTPECA d'après Rui Santos
  Complete project details at 
  https://RandomNerdTutorials.com/esp-now-one-to-many-esp32-esp8266/
  https://randomnerdtutorials.com/esp-now-esp32-arduino-ide/
  https://randomnerdtutorials.com/esp-now-esp8266-nodemcu-arduino-ide/  
  Permission is hereby granted, free of charge, to any person obtaining a copy
  of this software and associated documentation files.
  
  The above copyright notice and this permission notice shall be included in all
  copies or substantial portions of the Software.
****************************************************************************/

// Import required libraries
#ifdef ESP32
  #include <esp_now.h>
  #include <esp_wifi.h>
  #include <WiFi.h>
#else
  #include <espnow.h>
  #include <ESP8266WiFi.h>
#endif
#define DEBUG

/*--------------------------------------------------------------------------
 * RECUPERATION CHANNEL DE L'EMETTEUR
---------------------------------------------------------------------------*/
constexpr char WIFI_SSID[] = "... le nom de votre routeur ...";
int32_t getWiFiChannel(const char *ssid)
{
  if (int32_t n = WiFi.scanNetworks())
  {
    for (uint8_t i = 0; i < n; i++)
    {
      if (!strcmp(ssid, WiFi.SSID(i).c_str()))
      {
        return WiFi.channel(i);
      }
    }
  }
  return 0;
}

/*--------------------------------------------------------------------------
 * DONNEES PROTOCOLE ESP-NOW
---------------------------------------------------------------------------*/
//Structure to receive data : Must match the sender structure
typedef struct 
{
  byte   numApp;    // numéro appareil (de 1 à 8)
  char   typApp;    // type de l'appareil ('A', 'J', 'T', 'S')
  char   etaApp;    // état de l'appareil ( de '1' à '4')
  byte   noMot1;    // No du 1er moteur (de 1 à 8)
  byte   noMot2;    // No du 2me moteur (de 1 à 8)
} comm;

//Create a struct_message called myComm
comm myComm;
comm oldComm;

/*--------------------------------------------------------------------------
 * GESTION DES 74HC595
---------------------------------------------------------------------------*/
#ifdef ESP32
const    byte verrou     = 2;     // ST_CP (latch)
const    byte donnee     = 4;     // DS (data)
const    byte horloge    = 0;     // SH_CP (clock)
#else
const    byte verrou     = 13;    // ST_CP (latch) D7
const    byte donnee     = 14;    // DS (data)     D5
const    byte horloge    = 12;    // SH_CP (clock) D6
#endif

         uint16_t u_Data = 0b0101010101010101;     // data pour les 74hc595
const    uint16_t k_Msk  = 0b1100000000000000;     // Masque
const    uint16_t k_Dir  = 0b0100000000000000;     // Moteur direct
const    uint16_t k_Dev  = 0b1000000000000000;     // Moteur dévié

/*********************************************************************/
void ActionneAppareil() 
/*********************************************************************/
{
  char     typApp  = myComm.typApp;
  char     etaApp  = myComm.etaApp;
  byte     no_App  = myComm.numApp;
  
  int      no_Mot1 = myComm.noMot1 - 1;            // No 1er moteur
  int      no_Mot2 = myComm.noMot2 - 1;            // No 2me moteur
  int      nb_Dek1 = no_Mot1 + no_Mot1;
  int      nb_Dek2 = no_Mot2 < 0 ? 0 : no_Mot2 + no_Mot2;
  uint16_t w_Msk1  = ~(k_Msk >> nb_Dek1);
  uint16_t w_Msk2  = ~(k_Msk >> nb_Dek2);
  uint16_t w_Data;

  if (typApp == 'A')                                      // Aiguille simple
  {
    if (etaApp == '1')
    { 
      w_Data = (u_Data & w_Msk1) | (k_Dir >> nb_Dek1);
    }
    else
    { 
      w_Data = (u_Data & w_Msk1) | (k_Dev >> nb_Dek1);
    }
  }
  else if (typApp == 'S')                                 // Aiguille symétrique
  {
    if (etaApp == '1')
    { 
      w_Data = (u_Data & w_Msk1) | (k_Dir >> nb_Dek1);
    }
    else
    { 
      w_Data = (u_Data & w_Msk1) | (k_Dev >> nb_Dek1);
    }
  }
  else if (typApp == 'T')                                 // Aiguille triple
  {
     switch (etaApp) 
     {
       case '1' :  
          w_Data = (u_Data & w_Msk1) | (k_Dir >> nb_Dek1);
          w_Data = (w_Data & w_Msk2) | (k_Dir >> nb_Dek2);
          break; 
       case '2' :  
          w_Data = (u_Data & w_Msk1) | (k_Dev >> nb_Dek1);
          break; 
       default :          
          w_Data = (u_Data & w_Msk1) | (k_Dir >> nb_Dek1);
          w_Data = (w_Data & w_Msk2) | (k_Dev >> nb_Dek2);
          break; 
     }
  }
  else if (typApp == 'J')                                 // Traversée jonction
  {
     switch (etaApp) 
     {
       case '1' :  
          w_Data = (u_Data & w_Msk1) | (k_Dir >> nb_Dek1);
          w_Data = (w_Data & w_Msk2) | (k_Dir >> nb_Dek2);
          break; 
       case '2' :  
          w_Data = (u_Data & w_Msk1) | (k_Dir >> nb_Dek1);
          w_Data = (w_Data & w_Msk2) | (k_Dev >> nb_Dek2);
          break; 
       case '3' :  
          w_Data = (u_Data & w_Msk1) | (k_Dev >> nb_Dek1);
          w_Data = (w_Data & w_Msk2) | (k_Dev >> nb_Dek2);
          break; 
       default :          
          w_Data = (u_Data & w_Msk1) | (k_Dev >> nb_Dek1);
          w_Data = (w_Data & w_Msk2) | (k_Dir >> nb_Dek2);
          break; 
     }
  }
  else                                                   // rien faire
  {
  }

#ifdef DEBUG 
  (typApp == 'J') ? Serial.print ("Tjd") :  Serial.print ("Aig");
  Serial.print   ( no_App+1 );
  Serial.print   ( " : Av =" );
  Serial.print   ( u_Data, BIN);
  Serial.print   ( "  Ap =" );
  Serial.println ( w_Data, BIN);
#endif

  // Envoi données moteurs aux 74HC595
  //===================================================================
  u_Data = w_Data;
  digitalWrite(verrou, LOW);                  // mettre le verrou
  shiftOut(donnee, horloge, LSBFIRST, u_Data);
  shiftOut(donnee, horloge, LSBFIRST, (u_Data >> 8));
  digitalWrite(verrou, HIGH);                 // oter le verrou
  delay(500);
}

#ifdef ESP32
/*********************************************************************/
void OnDataRecv(const uint8_t * mac, const uint8_t *incomingData, int len) 
/*********************************************************************/
//callback function that will be executed when data is received
{
  memcpy(&myComm, incomingData, sizeof(myComm));
#ifdef DEBUG
//Serial.print("Octets reçus  : ");
//Serial.println(len);
  Serial.print("Nom  appareil : ");
  (myComm.typApp == 'J') ? Serial.print("Tjd") : Serial.print("Aig");
  Serial.print(String(myComm.numApp+1));
  Serial.print(" Type : ");
  Serial.print(String(myComm.typApp));
  Serial.print(" Etat : ");
  Serial.println(String(myComm.etaApp));
#endif
}
#else
/*********************************************************************/
void OnDataRecv(uint8_t * mac, uint8_t *incomingData, uint8_t len) 
/*********************************************************************/
{
  memcpy(&myComm, incomingData, sizeof(myComm));
#ifdef DEBUG
//Serial.print("Octets reçus  : ");
//Serial.println(len);
  Serial.print("Nom  appareil : ");
  (myComm.typApp == 'J') ? Serial.print("Tjd") : Serial.print("Aig");
  Serial.print(String(myComm.numApp+1));
  Serial.print(" Type : ");
  Serial.print(String(myComm.typApp));
  Serial.print(" Etat : ");
  Serial.println(String(myComm.etaApp));
#endif
}
#endif 
 
/*********************************************************************/
void setup() 
/*********************************************************************/
{
  //Initialize Serial Monitor for debugging purposes
  Serial.begin(115200);

  // Préparation des 74HC595
  //------------------------------------------------------------------
  pinMode(horloge, OUTPUT);
  pinMode(verrou,  OUTPUT);
  pinMode(donnee,  OUTPUT);
  
  myComm.numApp  = 1;
  myComm.typApp  = 'A';
  myComm.etaApp  = '1';
  oldComm.numApp = 1;
  oldComm.typApp = 'A';
  oldComm.etaApp = '1';

  // Connection à ESP-NOW
  //------------------------------------------------------------------
  WiFi.mode(WIFI_STA);
  int32_t channel = getWiFiChannel(WIFI_SSID);
  Serial.print   ( "SSID =  " );
  Serial.println ( WIFI_SSID );
  Serial.print   ( "WiFi channel : " );
  Serial.println ( WiFi.channel() );
  
#ifdef ESP32
  //WiFi.printDiag(Serial); // Uncomment to verify channel number before
  esp_wifi_set_promiscuous(true);
  esp_wifi_set_channel(channel, WIFI_SECOND_CHAN_NONE);
  esp_wifi_set_promiscuous(false);
  //WiFi.printDiag(Serial); // Uncomment to verify channel change after
#else
  //WiFi.printDiag(Serial); // Uncomment to verify channel number before
  wifi_promiscuous_enable(1);
  wifi_set_channel(channel);
  wifi_promiscuous_enable(0);
  //WiFi.printDiag(Serial); // Uncomment to verify channel change after
#endif


//WiFi.disconnect();
  Serial.println("");
  Serial.print("Initialisation ESP-NOW : ");
  
#ifdef ESP32
  if (esp_now_init() != ESP_OK) {
    Serial.println("Erreur");
    return;
  }
  Serial.println("Ok");
  // fonction callback
  esp_now_register_recv_cb(OnDataRecv);
#else
  if (esp_now_init() != 0) {
    Serial.println("Erreur");
    return;
  }
  Serial.println("Ok");
  // fonction callback
  esp_now_set_self_role(ESP_NOW_ROLE_CONTROLLER);
  esp_now_register_recv_cb(OnDataRecv);
#endif
  Serial.print ( "Mac[" );
  Serial.print ( WiFi.macAddress() );
  Serial.println("] prêt à recevoir.");

}
 
/*********************************************************************/
void loop() 
/*********************************************************************/
{
  if (myComm.numApp != oldComm.numApp ||
      myComm.typApp != oldComm.typApp ||
      myComm.etaApp != oldComm.etaApp)
  {
     ActionneAppareil();
     
     oldComm.numApp = myComm.numApp;
     oldComm.typApp = myComm.typApp;
     oldComm.etaApp = myComm.etaApp;
     delay(500);
  }
}
